import numpy as np
import matplotlib.pylab as plt
from sklearn.ensemble import RandomForestClassifier

from wildboar.ensemble import ShapeletForestClassifier
from wildboar.datasets import load_dataset
from wildboar.explain.counterfactual import PrototypeCounterfactual
from sklearn.neighbors import KNeighborsClassifier

x_train, x_test, y_train, y_test = load_dataset(
    "TwoLeadECG", repository="wildboar/ucr", merge_train_test=False
)
x_test_original = x_test
y_test_original = y_test

# clf = KNeighborsClassifier(n_neighbors=5, metric="euclidean")
clf = ShapeletForestClassifier(
    n_estimators=100, metric="scaled_euclidean", random_state=10
)
# clf = RandomForestClassifier()
clf.fit(x_train, y_train)

cf = PrototypeCounterfactual(
    x_train,
    y_train,
    metric="dtw",
    metric_params={"r": 0.1},
    method="nearest_shapelet",
    method_params={"min_shapelet_size": 0.1, "max_shapelet_size": 0.2},
    target=0.70,
    step_size=1,
    max_iter=500,
    n_prototypes=10,
    random_state=3,
)
cf.fit(clf)

y_pred = clf.predict(x_test)
class_ = clf.classes_[1]
print("Class: %s" % class_)
print("Pred: %r" % y_pred)
print(" - where %r " % (y_pred != class_).nonzero())
x_test = x_test[y_test != class_][:10]
y_test = y_test[y_test != class_][:10]


x_counterfactual, success = cf.transform(
    x_test, np.broadcast_to(class_, x_test.shape[0])
)


print(clf.predict_proba(x_counterfactual))
print(y_test)
x_test = x_test[success]
x_counterfactual = x_counterfactual[success]

fig, ax = plt.subplots(nrows=3)
ax[0].plot(x_counterfactual[4], c="red")
ax[0].plot(x_test[4], c="blue")
ax[1].plot(x_counterfactual[1], c="red")
ax[1].plot(x_test[1], c="blue")
ax[1].legend(["x'", "x"])
# ax[2].plot(np.mean(x_counterfactual, axis=0), "r--")
ax[2].plot(x_test_original[0], "b--")
ax[2].plot(x_test_original[1], "g--")
plt.show()



import numpy as np
from sklearn.metrics import roc_auc_score
from sklearn.model_selection import train_test_split

from wildboar.datasets import load_dataset
from wildboar.datasets.outlier import EmmottLabeler
from wildboar.ensemble import IsolationShapeletForest

x, y = load_dataset("SwedishLeaf", repository="wildboar/ucr", merge_train_test=True)
labeler = EmmottLabeler(
    n_outliers=0.05, difficulty=1, variation="tight", random_state=5
)
x, y = labeler.fit_transform(x, y)
print(np.unique(y, return_counts=True))
x_train, x_test, y_train, y_test = train_test_split(
    x, y, random_state=10, test_size=0.2, stratify=y
)

f = IsolationShapeletForest(
    random_state=10,
    n_estimators=100,
    contamination=0.1,
    metric="scaled_euclidean",
    n_jobs=-1,
    min_shapelet_size=0,
    max_shapelet_size=1,
)

f.fit(x_train, y_train)
y_pred = f.decision_function(x_test)
print("AUC=%f" % roc_auc_score(y_test, y_pred))


def plot_mds(x, y):
    import matplotlib.pylab as plt
    from sklearn.manifold import MDS

    m = MDS(n_components=2, random_state=1).fit_transform(x)
    m_normal = m[y == 1]
    m_outlier = m[y == -1]
    plt.scatter(m_normal[:, 0], m_normal[:, 1], c="blue")
    plt.scatter(m_outlier[:, 0], m_outlier[:, 1], c="red")
    plt.show()


plot_mds(x, y)



import numpy as np
from sklearn.model_selection import cross_validate, StratifiedKFold

from wildboar.datasets import load_dataset
from wildboar.ensemble import ShapeletForestClassifier, ExtraShapeletTreesClassifier

random_state = 1234

x, y = load_dataset("Beef")

classifiers = {
    "Shapelet forest": ShapeletForestClassifier(
        n_shapelets=10,
        metric="scaled_euclidean",
        n_jobs=-1,
        random_state=random_state,
    ),
    "Extra Shapelet Trees": ExtraShapeletTreesClassifier(
        metric="scaled_euclidean",
        n_jobs=-1,
        random_state=random_state,
    ),
}

for name, clf in classifiers.items():
    score = cross_validate(clf, x, y, scoring="roc_auc_ovo", n_jobs=1)
    print("Classifier: %s" % name)
    print(" - fit-time:   %.2f" % np.mean(score["fit_time"]))
    print(" - test-score: %.2f" % np.mean(score["test_score"]))



# This file is part of wildboar
#
# wildboar is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# wildboar is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
# Authors: Isak Samsten

import numpy as np
from sklearn.kernel_approximation import Nystroem
from sklearn.linear_model import LogisticRegression
from sklearn.utils.validation import check_is_fitted, check_random_state


class KernelLogisticRegression(LogisticRegression):
    """A simple kernel logistic implementation using a Nystroem kernel approximation

    Warnings
    --------
    This kernel method is not specialized for temporal classification.

    See Also
    --------
    wildboar.datasets.outlier.EmmottLabeler : Synthetic outlier dataset construction
    """

    def __init__(
        self,
        kernel=None,
        *,
        kernel_params=None,
        n_components=100,
        penalty="l2",
        dual=False,
        tol=1e-4,
        C=1.0,
        fit_intercept=True,
        intercept_scaling=1,
        class_weight=None,
        random_state=None,
        solver="lbfgs",
        max_iter=100,
        multi_class="auto",
        verbose=0,
        warm_start=False,
        n_jobs=None,
        l1_ratio=None
    ):
        """Create a new kernel logistic regression

        Parameters
        ----------
        kernel : str, optional
            The kernel function to use. See `sklearn.metrics.pairwise.kernel_metric` for kernels. The default kernel
            is 'rbf'.

        kernel_params : dict, optional
            Parameters to the kernel function.

        n_components : int, optional
            Number of features to construct
        """
        super().__init__(
            penalty=penalty,
            dual=dual,
            tol=tol,
            C=C,
            fit_intercept=fit_intercept,
            intercept_scaling=intercept_scaling,
            class_weight=class_weight,
            random_state=random_state,
            solver=solver,
            max_iter=max_iter,
            multi_class=multi_class,
            verbose=verbose,
            warm_start=warm_start,
            n_jobs=n_jobs,
            l1_ratio=l1_ratio,
        )
        self.kernel = kernel
        self.kernel_params = kernel_params
        self.n_components = n_components

    def fit(self, x, y, sample_weight=None):
        random_state = check_random_state(self.random_state)
        kernel = self.kernel or "rbf"
        n_components = min(x.shape[0], self.n_components)
        self.nystroem_ = Nystroem(
            kernel=kernel,
            kernel_params=self.kernel_params,
            n_components=n_components,
            random_state=random_state.randint(np.iinfo(np.int32).max),
        )
        self.nystroem_.fit(x)
        super().fit(self.nystroem_.transform(x), y, sample_weight=sample_weight)
        return self

    def decision_function(self, x):
        check_is_fitted(self)
        return super().decision_function(self.nystroem_.transform(x))



import numpy as np
import matplotlib.pylab as plt
from sklearn.ensemble import RandomForestClassifier
from sklearn.model_selection import train_test_split
from sklearn.neighbors import KNeighborsClassifier

from wildboar.ensemble import ShapeletForestClassifier
from wildboar.datasets import load_dataset
from wildboar.explain.counterfactual import counterfactuals

random_state = 1234
x, y = load_dataset("GunPoint")

x_train, x_test, y_train, y_test = train_test_split(
    x, y, test_size=0.2, random_state=random_state
)

classifiers = [
    (
        "Shapelet Forest Counterfactuals",
        ShapeletForestClassifier(
            metric="euclidean", random_state=random_state, n_estimators=100
        ),
    ),
    ("KNearest Counterfactuals", KNeighborsClassifier(metric="euclidean")),
    ("Prototype Counterfactuals", RandomForestClassifier(random_state=random_state)),
]

fig, ax = plt.subplots(nrows=3, sharex=True)
label = np.unique(y_train)[0]
for i, (name, clf) in enumerate(classifiers):
    clf.fit(x_train, y_train)
    x_test_sample = x_test[y_test != label]
    if isinstance(clf, RandomForestClassifier):
        kwargs = {"background_x": x_train, "background_y": y_train}
    else:
        kwargs = {}

    x_counterfactual, valid = counterfactuals(
        clf, x_test_sample, label, random_state=random_state, **kwargs
    )
    ax[i].set_title(name + ("(invalid)" if not valid[0] else ""))
    ax[i].plot(x_counterfactual[0], c="red")
    ax[i].plot(x_test[0], c="blue")

ax[-1].legend(["x'", "x"])
plt.savefig("../fig/counterfactuals.png")



# cython: language_level=3

# This file is part of wildboar
#
# wildboar is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# wildboar is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

# Authors: Isak Samsten
cimport numpy as np

from ..distance._distance cimport DistanceMeasure
from ..distance._distance cimport TSView, TSCopy

# TODO: include impurity score...
cdef struct SplitPoint:
    Py_ssize_t split_point
    double threshold
    TSView shapelet_info

cdef class Tree:
    cdef DistanceMeasure distance_measure

    cdef Py_ssize_t _max_depth
    cdef Py_ssize_t _capacity
    cdef Py_ssize_t _n_labels  # 1 for regression

    cdef Py_ssize_t _node_count
    cdef Py_ssize_t *_left
    cdef Py_ssize_t *_right
    cdef TSCopy ** _shapelets
    cdef double *_thresholds
    cdef double *_impurity
    cdef double *_values
    cdef double *_n_weighted_node_samples
    cdef Py_ssize_t *_n_node_samples

    cdef Py_ssize_t _increase_capacity(self) nogil except -1

    cdef Py_ssize_t add_leaf_node(self, Py_ssize_t parent, bint is_left, Py_ssize_t n_node_samples, double n_weighted_node_samples) nogil

    cdef void set_leaf_value(self, Py_ssize_t node_id, Py_ssize_t out_label, double out_value) nogil

    cdef Py_ssize_t add_branch_node(self, Py_ssize_t parent, bint is_left, Py_ssize_t n_node_samples, double n_weighted_node_samples,
                             TSCopy *shapelet, double threshold, double impurity) nogil

    cpdef np.ndarray predict(self, object X)

    cpdef np.ndarray apply(self, object X)

    cpdef np.ndarray decision_path(self, object X)



# This file is part of wildboar
#
# wildboar is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# wildboar is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
# Authors: Isak Samsten

import math

import numpy as np
from sklearn import model_selection
from sklearn.utils import check_random_state
from sklearn.utils._random import sample_without_replacement

from wildboar.model_selection._cv import RepeatedOutlierSplit

__all__ = [
    "train_test_split",
    "threshold_score",
    "RepeatedOutlierSplit",
]


def train_test_split(
    x, y, normal_class, test_size=0.2, anomalies_train_size=0.05, random_state=None
):
    """Training and testing split from classification dataset

    Parameters
    ----------
    x : array-like of shape (n_samples, n_timestep) or (n_samples, n_dim, n_timestep)
        Input data samples

    y : array-like of shape (n_samples,)
        Input class label

    normal_class : int
        Class label that should be considered as the normal class

    test_size : float
        Size of the test set

    anomalies_train_size : float
        Contamination of anomalies in the training dataset

    random_state : int or RandomState
        Psudo random state used for stable results.

    Returns
    -------
    x_train : array-like
        Training samples

    x_test : array-like
        Test samples

    y_train : array-like
        Training labels (either 1 or -1, where 1 denotes normal and -1 anomalous)

    y_test : array-like
        Test labels (either 1 or -1, where 1 denotes normal and -1 anomalous)

    Examples
    --------

    >>> from wildboar.datasets import load_two_lead_ecg
    >>> x, y = load_two_lead_ecg()
    >>> x_train, x_test, y_train, y_test = train_test_split(x, y, 1, test_size=0.2, anomalies_train_size=0.05)
    """
    random_state = check_random_state(random_state)
    normal = y == normal_class
    y = y.copy()
    y[normal] = 1
    y[~normal] = -1

    x_normal = x[np.where(y == 1)]
    x_anomalous = x[np.where(y == -1)]
    y_normal = y[np.where(y == 1)]
    y_anomalous = y[np.where(y == -1)]

    (xn_train, xn_test, yn_train, yn_test) = model_selection.train_test_split(
        x_normal, y_normal, test_size=test_size, random_state=random_state
    )

    n_sample = min(x_anomalous.shape[0], x_normal.shape[0])
    idx = sample_without_replacement(
        x_anomalous.shape[0], n_sample, random_state=random_state
    )
    n_training_anomalies = math.ceil(xn_train.shape[0] * anomalies_train_size)
    idx = idx[:n_training_anomalies]

    x_anomalous_train = x_anomalous[idx, :]
    y_anomalous_train = y_anomalous[idx]

    x_anomalous_test = np.delete(x_anomalous, idx, axis=0)
    y_anomalous_test = np.delete(y_anomalous, idx)

    x_train = np.vstack([xn_train, x_anomalous_train])
    y_train = np.hstack([yn_train, y_anomalous_train])
    x_test = np.vstack([xn_test, x_anomalous_test])
    y_test = np.hstack([yn_test, y_anomalous_test])
    return x_train, x_test, y_train, y_test


def threshold_score(y_true, score, score_f):
    """Compute the performance of using the i:th score

    The scores are typically computed using an outlier detection algorithm

    Parameters
    ----------
    y_true : array-like
        The true labels

    score : array-like
        The scores

    score_f : callable
        Function for estimating the performance of the i:th scoring

    Returns
    -------
    score : ndarray
        performance for each score as threshold

    See Also
    --------
    wildboar.ensemble.IsolationShapeletForest : an isolation forest for time series

    Examples
    --------
    Setting the offset that maximizes balanced accuracy of a shapelet isolation forest

    >>> from wildboar.ensemble import IsolationShapeletForest
    >>> from wildboar.datasets import load_two_lead_ecg
    >>> from sklearn.metrics import balanced_accuracy_score
    >>> x, y = load_two_lead_ecg()
    >>> x_train, x_test, y_train, y_test = train_test_split(x, y, 1, test_size=0.2, anomalies_train_size=0.05)
    >>> f = IsolationShapeletForest()
    >>> f.fit(x_train)
    >>> scores = f.score_samples(x_train)
    >>> perf = threshold_score(y_train, scores, balanced_accuracy_score)
    >>> f.offset_ = score[np.argmax(perf)]
    """
    ba_score = np.empty(score.shape[0], dtype=np.float)
    score_copy = np.empty(score.shape[0], dtype=np.float)
    is_inlier = np.ones(score.shape[0])
    for i in range(score.shape[0]):
        score_copy[:] = score
        is_inlier[:] = 1
        is_inlier[score_copy - score[i] < 0] = -1
        ba_score[i] = score_f(y_true, is_inlier)
    return ba_score



import os
from .tutorial import PLOT_DICT as TUTORIAL
from .counterfactual import PLOT_DICT as COUNTERFACTUALS


def build_all(output_path, plot_dict, subdir=None, output_ext=".png"):
    if subdir:
        output_path = os.path.join(output_path, subdir)

    if not os.path.exists(output_path):
        os.mkdir(output_path)

    for name, plotter in plot_dict.items():
        print("Plotting: %s" % name)
        fig = plotter()
        fig.savefig(os.path.join(output_path, name + output_ext))



# cython: language_level=3

# This file is part of wildboar
#
# wildboar is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# wildboar is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
# License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see
# <http://www.gnu.org/licenses/>.
#
# Authors: Isak Samsten

cimport numpy as np


cdef struct TSView:
    Py_ssize_t index  # the index of the shapelet sample
    Py_ssize_t start  # the start position
    Py_ssize_t length  # the length of the shapelet
    Py_ssize_t dim  # the dimension of the shapelet
    double mean  # the mean of the shapelet
    double std  # the stanard devision
    void *extra


cdef struct TSDatabase:
    Py_ssize_t n_samples  # the number of samples
    Py_ssize_t n_timestep  # the number of timesteps
    Py_ssize_t n_dims

    double *data  # the data
    Py_ssize_t sample_stride  # the stride for samples
    Py_ssize_t timestep_stride  # the `feature` stride
    Py_ssize_t dim_stride  # the dimension stride


cdef struct TSCopy:
    Py_ssize_t length
    Py_ssize_t dim
    double mean
    double std
    int ts_index
    int ts_start
    double *data
    void *extra


cdef class DistanceMeasure:
    cdef Py_ssize_t n_timestep


    cdef int init(self, TSDatabase *td) nogil


    cdef int init_ts_view(
        self,
        TSDatabase *td,
        TSView *ts_view,
        Py_ssize_t index,
        Py_ssize_t start,
        Py_ssize_t length,
        Py_ssize_t dim,
    ) nogil


    cdef int init_ts_copy_from_ndarray(
        self,
        TSCopy *ts_copy,
        np.ndarray arr,
        Py_ssize_t dim,
    )


    cdef int init_ts_copy(
        self,
        TSCopy *ts_copy,
        TSView *s,
        TSDatabase *td,
    ) nogil


    cdef int ts_copy_sub_matches(
        self,
        TSCopy *ts_copy,
        TSDatabase *td,
        Py_ssize_t t_index,
        double threshold,
        Py_ssize_t** matches,
        double** distances,
        Py_ssize_t *n_matches,
    ) nogil except -1


    cdef double ts_copy_sub_distance(
        self,
        TSCopy *ts_copy,
        TSDatabase *td,
        Py_ssize_t t_index,
        Py_ssize_t *return_index= *,
    ) nogil


    cdef double ts_view_sub_distance(
        self,
        TSView *ts_view,
        TSDatabase *td,
        Py_ssize_t t_index,
    ) nogil


    cdef void ts_view_sub_distances(
        self,
        TSView *ts_view,
        TSDatabase *td,
        Py_ssize_t *samples,
        double *distances,
        Py_ssize_t n_samples,
    ) nogil


    cdef double ts_copy_distance(
        self,
        TSCopy *ts_copy,
        TSDatabase *td,
        Py_ssize_t t_index,
    ) nogil


    cdef bint support_unaligned(self) nogil


cdef class ScaledDistanceMeasure(DistanceMeasure):
    pass


cdef TSDatabase ts_database_new(np.ndarray X)


cdef void ts_view_init(TSView *s) nogil


cdef void ts_view_free(TSView *shapelet_info) nogil


cdef int ts_copy_init(
    TSCopy *ts_copy,
    Py_ssize_t dim,
    Py_ssize_t length,
    double mean,
    double std,
) nogil


cdef void ts_copy_free(TSCopy *shapelet) nogil


cdef DistanceMeasure new_distance_measure(
    TSDatabase *td,
    object metric,
    dict metric_params=*,
)



import math

import numpy as np

from ._dtw_distance import _dtw_alignment
from ._dtw_distance import _dtw_distance
from ._dtw_distance import _dtw_pairwise_distance
from ._dtw_distance import _dtw_envelop
from ._dtw_distance import _dtw_lb_keogh

from .._utils import check_array_fast

__all__ = [
    "dtw_alignment",
    "dtw_distance",
    "dtw_mapping",
    "dtw_envelop",
    "dtw_pairwise_distance",
    "dtw_lb_keogh",
]


def _compute_warp_size(x_size, r, *, y_size=0):
    """Compute the warping window from array sizes and r

    Notes
    -----
    This function ensures that r is always in the range [1, max(x_size, y_size) - 1]
    """
    x_size = max(x_size, y_size)
    if isinstance(r, float):
        if not 0.0 <= r <= 1.0:
            raise ValueError("r should be in [0, 1], got %r" % r)
        return max(min(math.floor(x_size * r), x_size - 1), 1)
    elif isinstance(r, int) and 0 <= r < x_size:
        return max(1, r)
    else:
        raise ValueError("invalid warping window, got %r" % r)


def dtw_alignment(x, y, r=1.0, out=None):
    """Compute the Dynamic time warping alignment matrix

    Parameters
    ----------
    x : array-like of shape (x_timestep,)
        The first time series

    y : array-like of shape (y_timestep,)
        The second time series

    r : int or float, optional
        The warping window

        - if float in [0, 1] a fraction of max(x_timestep, y_timestep)
        - if int the exact warping window (max(1, r))

    out : array-like of shape (x_timestep, y_timestep), optional
        Store the warping path in this array.

    Returns
    -------
    alignment : ndarray of shape (x_timestep, y_timestep)
        The dynamic time warping alignment matrix

    Notes
    --------
    If only the distance between two series is required use `dtw_distance` instead

    See Also
    --------
    dtw_distance : compute the dtw distance
    """
    x = np.asarray(x)
    y = np.asarray(y)
    warp_size = _compute_warp_size(x.shape[0], r, y_size=y.shape[0])
    return _dtw_alignment(x, y, warp_size, out)


def dtw_distance(x, y, r=1.0, scale=False):
    """Compute the dynamic time warping distance

    Parameters
    ----------
    x : array-like of shape (x_timestep, )
        The first time series

    y : array-like of shape (y_timestep, )
        The second time series

    r : int or float, optional
        The warping window

        - if float in [0, 1] a fraction of max(x_timestep, y_timestep)
        - if int the exact warping window (max(1, r))

    scale : bool, optional
        If True, x and y are standardized before calculation

    Returns
    -------
    distance : float
        The dynamic time warping distance

    See Also
    --------
    dtw_alignment : compute the dtw alignment matrix
    """
    x = np.asarray(x)
    y = np.asarray(y)
    warp_size = _compute_warp_size(x.shape[0], r, y_size=y.shape[0])
    return _dtw_distance(x, y, warp_size, scale=scale)


def dtw_pairwise_distance(x, r=1.0):
    """Compute the distance between all pairs of rows

    Parameters
    ----------
    x : array-like of shape (n_samples, n_timestep)
        An array of samples

    r : float or int, optional
        The size of the warping window

        - if float in [0, 1] a fraction of x_timestep
        - if int the exact warping window (max(1, r))

    Returns
    -------
    distances : ndarray of shape (n_samples, n_samples)
        The distance between pairs of rows
    """
    x = np.asarray(x)
    if not x.ndim == 2:
        raise ValueError("requires a 2d-array")

    warp_size = _compute_warp_size(x.shape[1], r)
    return _dtw_pairwise_distance(x, warp_size)


def dtw_envelop(x, r=1.0):
    """Compute the envelop for LB_keogh

    Parameters
    ----------
    x : array-like of shape (x_timestep,)
        The time series

    r : float or int, optional
        The size of the warping window

        - if float in [0, 1] a fraction of x_timestep
        - if int the exact warping window (max(1, r))

    Returns
    -------
    lower : ndarray of shape (x_timestep,)
        The min value of the envelop

    upper : ndarray of shape (x_timestep,)
        The max value of the envelop

    References
    ----------
    Keogh, E. (2002).
        Exact indexing of dynamic time warping.
        In 28th International Conference on Very Large Data Bases.
    """
    x = np.asarray(x)
    warp_size = _compute_warp_size(x.shape[0], r)
    return _dtw_envelop(x, warp_size)


def dtw_lb_keogh(x, y=None, *, lower=None, upper=None, r=1.0):
    """The LB_keogh lower bound

    Parameters
    ----------
    x : array-like of shape (x_timestep,)
        The first time series

    y : array-like of shape (x_timestep,), optional
        The second time series (same size as x)

    lower : ndarray of shape (x_timestep,), optional
        The min value of the envelop

    upper : ndarray of shape (x_timestep,), optional
        The max value of the envelop

    r : float or int, optional
        The size of the warping window

        - if float in [0, 1] a fraction of x_timestep
        - if int the exact warping window (max(1, r))

    Returns
    -------
    min_dist : float
        The cumulative minimum distance.

    lb_keogh : ndarray of shape (x_timestep,),
        The lower bound at each time step

    Notes
    -----
    - if y=None, both lower and upper must be given
    - if y is given, lower and upper are ignored
    - if lower and upper is given and y=None, r is ignored

    References
    ----------
    Keogh, E. (2002).
        Exact indexing of dynamic time warping.
        In 28th International Conference on Very Large Data Bases.
    """
    x = np.asarray(x)
    if y is not None:
        y = np.asarray(y)
        if y.shape[0] != x.shape[0]:
            raise ValueError("invalid shape, got (%d, %d)" % (x.shape[0], y.shape[0]))
        lower, upper = dtw_envelop(y, r)
    elif lower is None or upper is None:
        raise ValueError("both y, lower and upper can't be None")

    if lower.shape[0] != upper.shape[0] or lower.shape[0] != x.shape[0]:
        raise ValueError(
            "invalid shape for lower (%d), upper (%d) and x (%d)"
            % (lower.shape[0], upper.shape[0], x.shape[0])
        )
    warp_size = _compute_warp_size(x.shape[0], r)
    return _dtw_lb_keogh(x, lower, upper, warp_size)


def dtw_mapping(x=None, y=None, *, alignment=None, r=1, return_index=False):
    """Compute the optimal warping path between two series or from a given
    alignment matrix

    Parameters
    ----------
    x : array-like of shape (x_timestep,), optional
        The first time series

    y : array-like of shape (y_timestep,), optional
        The second time series

    alignment : ndarray of shape (x_timestep, y_timestep), optional
        Precomputed alignment

    r : float or int, optional
        - if float, the warping path is a fraction of

    return_index : bool, optional
        Return the indices of the warping path

    Returns
    -------
    indicator : ndarray of shape (x_timestep, y_timestep)
        Boolean array with the dtw path

    (x_indices, y_indices) : tuple, optional
        The indices of the first and second dimension of
        the optimal alignment path.

    Notes
    -----
    - either x and y or alignment must be provided
    - if alignment is given x and y are ignored
    - if alignment is given r is ignored
    """
    if alignment is None:
        if x is None or y is None:
            raise ValueError("if alignment=None, neither x or y can be None")
        alignment = dtw_alignment(x, y, r=r)

    indicator = np.zeros(alignment.shape).astype(bool)
    i = alignment.shape[0] - 1
    j = alignment.shape[1] - 1
    while i > 0 or j > 0:
        indicator[i, j] = True
        option_diag = alignment[i - 1, j - 1] if i > 0 and j > 0 else np.inf
        option_up = alignment[i - 1, j] if i > 0 else np.inf
        option_left = alignment[i, j - 1] if j > 0 else np.inf
        move = np.argmin([option_diag, option_up, option_left])
        if move == 0:
            i -= 1
            j -= 1
        elif move == 1:
            i -= 1
        else:
            j -= 1
    indicator[0, 0] = True
    if return_index:
        return indicator, indicator.nonzero()
    else:
        return indicator



# This file is part of wildboar
#
# wildboar is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# wildboar is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
# Authors: Isak Samsten
import abc
import math
from copy import deepcopy
from functools import partial

import numpy as np
from sklearn.neighbors import NearestNeighbors
from sklearn.utils.validation import check_is_fitted, check_array, check_random_state

from wildboar.distance.dtw import dtw_mapping, dtw_distance
from wildboar.distance import distance

from .base import BaseCounterfactual


class PrototypeCounterfactual(BaseCounterfactual):
    """Model agnostic approach for constructing counterfactual explanations

    Attributes
    ----------

    estimator_ : object
        The estimator for which counterfactuals are computed

    classes_ : ndarray
        The classes

    partitions_ : dict
        Dictionary of classes and PrototypeSampler

    target_ : TargetEvaluator
        The target evaluator

    References
    ----------

    Samsten, Isak (2020).
        Model agnostic time series counterfactuals
    """

    def __init__(
        self,
        background_x,
        background_y,
        *,
        metric="euclidean",
        metric_params=None,
        max_iter=100,
        step_size=0.1,
        n_prototypes="auto",
        target="auto",
        method="sample",
        method_params=None,
        random_state=None
    ):
        """Crate a new model agnostic counterfactual explainer.

        Parameters
        ----------
        background_x : array-like of shape (n_samples, n_timestep)
            The background data from which prototypes are sampled

        background_y : array-like of shape (n_samples,)
            The background label from which prototypes are sampled

        metric : {'euclidean', 'dtw'}, optional
            The metric used to move the samples

        metric_params : dict, optional
            Optional parameters to the metric

            If 'dtw':

                r : int or float, optional
                    The warping window size

        max_iter : int, optional
            The maximum number of iterations

        step_size : float, optional
            The step size when moving samples toward class prototypes

        n_prototypes : int, float or str, optional
            The number of initial prototypes to sample from

        target : float or str, optional
            The target evaluation of counterfactuals:

            - if 'auto' the counterfactual prediction must return the correct
              label
            - if float, the counterfactual prediction probability must
              exceed target value

        method : {'sample', 'shapelet', 'nearest', 'nearest_shapelet'}, optional
            Method for selecting prototypes

            - if 'sample' a prototype is sampled among the initial prototypes
            - if 'shapelet' a prototype shapelet is sampled among the initial
              prototypes
            - if 'nearest' a prototype is sampled from the closest n prototypes
            - if 'nearest_shapelet' a prototype shapelet is sampled from the
              closest n prototypes

        method_params : dict, optional
            Additional parameters to the method

            If 'shapelet' or 'nearest_shapelet'

                min_shapelet_size : float, optional
                    Minimum shapelet size.

                max_shapelet_size : float, optional
                    Maximum shapelet size.

        random_state : RandomState or int, optional
            Pseudo-random number for consistency between different runs
        """
        self.background_x = background_x
        self.background_y = background_y
        self.random_state = random_state
        self.metric = metric
        self.metric_params = metric_params
        self.max_iter = max_iter
        self.step_size = step_size
        self.n_prototypes = n_prototypes
        self.method = method
        self.method_params = method_params
        self.target = target

    def fit(self, estimator):
        check_is_fitted(estimator)
        if self.background_x is None or self.background_y is None:
            raise ValueError("background data are required.")

        x = check_array(self.background_x)
        y = check_array(self.background_y, ensure_2d=False)
        if len(y) != x.shape[0]:
            raise ValueError(
                "Number of labels={} does not match "
                "number of samples={}".format(len(y), x.shape[0])
            )
        random_state = check_random_state(self.random_state)
        metric_params = self.metric_params or {}
        method_params = self.method_params or {}
        if self.metric in _METRIC_TRANSFORM:
            metric = _METRIC_TRANSFORM[self.metric](self.step_size, **metric_params)
        else:
            raise ValueError("metric (%s) is not supported" % self.metric)

        if self.method in _PROTOTYPE_SAMPLER:
            sampler = _PROTOTYPE_SAMPLER[self.method]
        else:
            raise ValueError("method (%s) is not supported" % self.method)

        self.estimator_ = deepcopy(estimator)
        self.classes_ = np.unique(self.background_y)
        if self.target == "auto":
            self.target_ = PredictEvaluator(self.estimator_)
        else:
            if not 0 < self.target <= 1.0:
                raise ValueError("target must be in (0, 1], got %r" % self.target)
            self.target_ = ProbabilityEvaluator(self.estimator_, self.target)

        self.partitions_ = {}
        for c in self.classes_:
            x_partition = x[y == c]
            if self.n_prototypes == "auto":
                n_prototypes = x_partition.shape[0]
            elif isinstance(self.n_prototypes, int):
                n_prototypes = max(1, min(self.n_prototypes, x_partition.shape[0]))
            elif isinstance(self.n_prototypes, float):
                if not 0.0 < self.n_prototypes <= 1.0:
                    raise ValueError("n_prototypes")
                n_prototypes = math.ceil(self.n_prototypes * x_partition.shape[0])
            else:
                raise ValueError("n_prototypes (%r) not supported" % self.n_prototypes)

            self.partitions_[c] = sampler(
                x_partition, c, n_prototypes, metric, random_state, **method_params
            )

    def transform(self, x, y):
        x = check_array(x)
        y = check_array(y, ensure_2d=False)
        if len(y) != x.shape[0]:
            raise ValueError(
                "Number of labels={} does not match "
                "number of samples={}".format(len(y), x.shape[0])
            )
        n_samples = x.shape[0]
        counterfactuals = np.empty(x.shape, dtype=x.dtype)
        success = np.ones(x.shape[0]).astype(bool)
        for i in range(n_samples):
            counterfactual = self._transform_sample(x[i], y[i])
            if counterfactual is not None:
                counterfactuals[i] = counterfactual
            else:
                success[i] = False
        return counterfactuals, success

    def _transform_sample(self, x, y):
        sampler = self.partitions_[y]
        o = x.copy()
        n_iter = 0
        while not self.target_.is_counterfactual(o, y) and n_iter < self.max_iter:
            o = sampler.sample_move(o)
            n_iter += 1

        if n_iter > self.max_iter:
            return None
        else:
            return o


class TargetEvaluator(abc.ABC):
    """Evaluate if a sample is a counterfactual"""

    def __init__(self, estimator):
        """Construct a new evaluator

        Parameters
        ----------
        estimator : object
            The estimator
        """
        self.estimator = estimator

    def is_counterfactual(self, x, y):
        """Return true if x is a counterfactual of label y

        Parameters
        ----------
        x : ndarray of shape (n_samples, n_timestep) or (n_timestep,)
            The counterfactual sample

        y : object
            The counterfactual label

        Returns
        -------
        bool : true if counterfactual
        """
        return self._is_counterfactual(x.reshape(1, -1), y)

    @abc.abstractmethod
    def _is_counterfactual(self, x, y):
        pass


class PredictEvaluator(TargetEvaluator):
    """Evaluate if a counterfactual is predicted as y"""

    def _is_counterfactual(self, x, y):
        return self.estimator.predict(x) == y


class ProbabilityEvaluator(TargetEvaluator):
    """Evaluate if the probability of a counterfactual is at least a given constant"""

    def __init__(self, estimator, probability=0.5):
        """Construct a new evaluator

        Parameters
        ----------
        estimator : object
            The estimator

        probability : float
            The minimum probability of the predicted label
        """
        super().__init__(estimator)
        self.probability = probability

    def _is_counterfactual(self, x, y):
        if not hasattr(self.estimator, "predict_proba"):
            raise ValueError("estimator must support predict_proba")

        y_pred = self.estimator.predict_proba(x)
        y_idx = (self.estimator.classes_ == y).nonzero()[0][0]
        y_prob = y_pred[0, y_idx]
        return y_prob > self.probability


class PrototypeSampler(abc.ABC):
    def __init__(self, x, y, n_prototypes, metric_transform, random_state):
        """Sample and refine counterfactuals

        Parameters
        ----------
        x : ndarray of shape (n_samples, n_timestep)
            The data samples labeled as y

        y : object
            The label of the samples in x

        n_prototypes : int
            The number of prototypes in the initial sample

        metric_transform : MetricTransform
            The metric transformer.

        random_state : RandomState
            The random number generator.
        """
        self.random_state = random_state
        self.x = x
        self.y = y
        self.metric_transform = metric_transform
        self.n_prototypes = n_prototypes
        prototype_indices = np.arange(x.shape[0])
        self.random_state.shuffle(prototype_indices)
        self.prototype_indices = prototype_indices[:n_prototypes]

    @property
    def _random_index(self):
        """Return a random index in the initial prototype sample

        Returns
        -------
        int : an index
        """
        return self.prototype_indices[
            self.random_state.randint(self.prototype_indices.shape[0])
        ]

    def _random_indices(self, n):
        """Return n random indices from the initial prototype sample

        Parameters
        ----------
        n : int
            The number of idices to return

        Returns
        -------
        ndarray : indices
        """
        n = min(self.n_prototypes - 1, max(1, n))
        self.random_state.shuffle(self.prototype_indices)
        return self.prototype_indices[:n]

    @abc.abstractmethod
    def sample(self, o):
        """Sample an example

        Parameters
        ----------
        o : ndarray of shape (n_timestep,)
            The current counterfactual sample

        Returns
        -------
        prototype : ndarray of shape (n_timestep,)
            A prototype of the counterfactual label
        """
        pass

    def move(self, o, p):
        """Move the current counterfactual toward the prototype

        Parameters
        ----------
        o : ndarray of shape (n_timestep,)
            The current counterfactual sample

        p : ndarray of shape (n_timestep,)
            The prototype of the counterfactual label

        Returns
        -------
        new_counterfactual : ndarray of shape (n_timestep,)
            The new counterfactual moved towards the prototype
        """
        return self.metric_transform.move(o, p)

    def sample_move(self, o):
        """Sampla a prototype and move the counterfactual towards the prototype

        Parameters
        ----------
        o : ndarray of shape (n_timestep,)
            The current counterfactual sample

        Returns
        -------
        new_counterfactual : ndarray of shape (n_timestep,)
            The new counterfactual moved towards the prototype
        """
        p = self.sample(o)
        return self.move(o, p)


class UniformPrototypeSampler(PrototypeSampler):
    """Sample a prototype uniformly at random from the initial prototype sample"""

    def sample(self, _o):
        return self.x[self._random_index]


class KNearestPrototypeSampler(PrototypeSampler):
    """Sample a prototype among the samples closest to the current counterfactual"""

    def __init__(self, x, y, n_prototypes, metric_transform, random_state):
        super().__init__(x, y, n_prototypes, metric_transform, random_state)
        if isinstance(self.metric_transform, EuclideanTransform):
            self.nearest_neighbors = NearestNeighbors(
                metric="euclidean", n_neighbors=n_prototypes
            )
        elif isinstance(self.metric_transform, DynamicTimeWarpTransform):
            dtw = partial(dtw_distance, r=self.metric_transform.r)
            self.nearest_neighbors = NearestNeighbors(
                metric=dtw, n_neighbors=n_prototypes
            )
        else:
            raise ValueError("unsupported metric")

        self.nearest_neighbors.fit(x)

    def nearest_index(self, o):
        """Return the index of the closest sample

        Parameters
        ----------
        o : ndarray of shape (n_timestep,)
            The current counterfactual sample

        Returns
        -------
        int : an index
        """
        nearest = self.nearest_neighbors.kneighbors(
            o.reshape(1, -1), return_distance=False
        ).reshape(-1)
        return nearest[self.random_state.randint(nearest.shape[0])]

    def sample(self, o):
        return self.x[self.nearest_index(o)]


class ShapeletPrototypeSampler(PrototypeSampler):
    """Sample shapelet prototypes"""

    def __init__(
        self,
        x,
        y,
        n_prototypes,
        metric_transform,
        random_state,
        min_shapelet_size=0,
        max_shapelet_size=1,
    ):
        """Sample shapelet

        Parameters
        ----------
        x : ndarray of shape (n_samples, n_timestep)
            The data samples

        y : object
            The label of the samples in x

        metric_transform : MetricTransform
            The metric transformer.

        random_state : RandomState
            The random number generator.

        min_shapelet_size : float
            The minimum shapelet size

        max_shapelet_size : float
            The maximum shapelet size
        """
        super().__init__(x, y, n_prototypes, metric_transform, random_state)
        self.min_shapelet_size = min_shapelet_size
        self.max_shapelet_size = max_shapelet_size

    def sample_shapelet(self, p):
        """Sample a shapelet from x

        Parameters
        ----------
        p : ndarray of shape (n_timestep,)
             The prototype sample

        Returns
        -------
        shapelet : ndarray
            A shapelet
        """
        n_timestep = self.x.shape[-1]
        min_shapelet_size = max(2, int(n_timestep * self.min_shapelet_size))
        max_shapelet_size = int(n_timestep * self.max_shapelet_size)
        shapelet_length = self.random_state.randint(
            min_shapelet_size, max_shapelet_size
        )
        start_index = self.random_state.randint(0, n_timestep - shapelet_length)
        return p[start_index : (start_index + shapelet_length)]

    def sample(self, _o):
        return self.sample_shapelet(self.x[self._random_index])

    def move(self, o, p):
        """Move the best matching shapelet of the  counterfactual sample towards
        the shapelet prototype

        Parameters
        ----------
        o : ndarray of shape (n_timestep,)
            The counterfactual sample

        p : ndarray
            The prototype shapelet

        Returns
        -------
        new_counterfactual : ndarray of shape (n_timestep,)
            The new counterfactual moved towards the prototype
        """
        if isinstance(self.metric_transform, EuclideanTransform):
            metric = "euclidean"
            metric_params = {}
        elif isinstance(self.metric_transform, DynamicTimeWarpTransform):
            metric = "dtw"
            metric_params = {"r": self.metric_transform.r}
        else:
            raise ValueError("unsupported metric")

        # Find the best matching position in
        min_dist, best_match = distance(
            p, o, metric=metric, metric_params=metric_params, return_index=True
        )
        o[best_match : best_match + p.shape[0]] = self.metric_transform.move(
            o[best_match : best_match + p.shape[0]], p
        )

        return o


class KNearestShapeletPrototypeSampler(PrototypeSampler):
    """Combines the KNearestPrototypeSample and the ShapeletPrototypeSampler
    such that prototype samples are sampled among the nearest neighbors of the
    counterfactual
    """

    def __init__(
        self,
        x,
        y,
        n_prototypes,
        metric_transform,
        random_state,
        min_shapelet_size=0,
        max_shapelet_size=1,
    ):
        super().__init__(x, y, n_prototypes, metric_transform, random_state)
        self.nearest_sampler = KNearestPrototypeSampler(
            x, y, n_prototypes, metric_transform, random_state
        )
        self.shapelet_sampler = ShapeletPrototypeSampler(
            x,
            y,
            n_prototypes,
            metric_transform,
            random_state,
            min_shapelet_size,
            max_shapelet_size,
        )

    def sample(self, o):
        p = self.nearest_sampler.nearest_index(o)
        return self.shapelet_sampler.sample_shapelet(self.x[p])

    def move(self, o, p):
        return self.shapelet_sampler.move(o, p)


class MetricTransform(abc.ABC):
    """Move a time series towards a prototype"""

    def __init__(self, gamma):
        """Construct a new transformer

        Parameters
        ----------
        gamma : float
            The strength of the move, where values close to 0 means that the
            sample is moved less and values closer to 1 mean that the sample
            is moved more.
        """
        if not 0.0 < gamma <= 1.0:
            raise ValueError("gamma must be in (0, 1], got %r" % gamma)
        self.gamma = gamma

    @abc.abstractmethod
    def move(self, o, p):
        """Move the sample o towards p

        Parameters
        ----------
        o : ndarray of shape (n_timestep,)
            An array
        p : ndarray of shape (n_timestep,)
            An array

        Returns
        -------
        ndarray : an array
        """
        pass


class EuclideanTransform(MetricTransform):
    """Transform a sample by moving it closer in euclidean space"""

    def move(self, o, p):
        return ((1 - self.gamma) * o) + self.gamma * p


class DynamicTimeWarpTransform(MetricTransform):
    """Transform a sample by moving it closer using the dtw alignment matrix"""

    def __init__(self, gamma, r=1.0):
        super().__init__(gamma)
        self.r = r

    def move(self, o, p):
        indicator, indices = dtw_mapping(o, p, r=self.r, return_index=True)
        result = o * (1 - self.gamma)
        weight = np.ones(o.shape[0]) * (1 - self.gamma)
        for i, j in zip(*indices):
            result[j] += self.gamma * p[j]
            weight[j] += self.gamma
        return result / weight


_METRIC_TRANSFORM = {
    "euclidean": EuclideanTransform,
    "dtw": DynamicTimeWarpTransform,
}

_PROTOTYPE_SAMPLER = {
    "sample": UniformPrototypeSampler,
    "nearest": KNearestPrototypeSampler,
    "shapelet": ShapeletPrototypeSampler,
    "nearest_shapelet": KNearestShapeletPrototypeSampler,
}



import numpy as np
import matplotlib.pylab as plt

from sklearn.neighbors import KNeighborsClassifier
from sklearn.model_selection import train_test_split

from wildboar.datasets import load_two_lead_ecg
from wildboar.ensemble import ShapeletForestClassifier
from wildboar.explain.counterfactual import counterfactuals


def counterfactuals_plot(estimator, method="infer"):
    x, y = load_two_lead_ecg()
    x_train, x_test, y_train, y_test = train_test_split(
        x, y, test_size=0.1, random_state=123
    )
    print("fitting the estimator %r" % estimator)
    estimator.fit(x_train, y_train)

    x_test = x_test[y_test == 2.0]

    print("computing %d counterfactuals" % x_test.shape[0])
    x_counterfactuals, success, score = counterfactuals(
        estimator, x_test, 1.0, scoring="euclidean", method=method, random_state=123
    )

    x_test = x_test[success]
    x_counterfactuals = x_counterfactuals[success]
    i = np.argsort(score[success])[:2]
    x_counterfactuals = x_counterfactuals[i, :]
    x_test = x_test[i, :]

    fig, ax = plt.subplots(ncols=1, nrows=2, figsize=(6, 2.5))
    ax[0].plot(x_test[0, :], c="red")
    ax[0].plot(x_counterfactuals[0, :], c="blue")
    ax[1].plot(x_test[1, :], c="red")
    ax[1].plot(x_counterfactuals[1, :], c="blue")
    ax[1].legend(["y=abnormal", "y*=normal"])
    return fig


def counterfactuals_nn():
    return counterfactuals_plot(KNeighborsClassifier(n_neighbors=5, metric="euclidean"))


def counterfactuals_sf():
    return counterfactuals_plot(
        ShapeletForestClassifier(random_state=123, n_jobs=-1, metric="euclidean")
    )


PLOT_DICT = {
    "counterfactuals_nn": counterfactuals_nn,
    "counterfactuals_sf": counterfactuals_sf,
}



# -*- coding: utf-8 -*-

import os
from setuptools import setup
from setuptools.extension import Extension


# Lazy loading
class np_include_dirs(str):
    def __str__(self):
        import numpy as np

        return np.get_include()


BUILD_TYPE = os.getenv("WILDBOAR_BUILD") or "default"

BUILD_ARGS = {
    "default": {
        "extra_compile_args": ["-O2"],
        "extra_link_args": [],
        "libraries": [],
    },
    "optimized": {
        "extra_compile_args": [
            "-O2",
            "-march=native",
            "-msse",
            "-msse2",
            "-mfma",
            "-mfpmath=sse",
        ],
        "extra_link_args": [],
        "libraries": [],
    },
}

if __name__ == "__main__":
    from Cython.Build import cythonize

    include_dirs = [np_include_dirs()]

    build_args = BUILD_ARGS.get(BUILD_TYPE)
    if build_args is None:
        raise RuntimeError("%s is not a valid build type" % BUILD_TYPE)

    ext_modules = [
        Extension(
            "wildboar._utils",
            sources=["src/wildboar/_utils.pyx"],
            include_dirs=include_dirs,
            **build_args,
        ),
        Extension(
            "wildboar.distance._distance",
            sources=["src/wildboar/distance/_distance.pyx"],
            include_dirs=include_dirs,
            **build_args,
        ),
        Extension(
            "wildboar.distance._euclidean_distance",
            sources=["src/wildboar/distance/_euclidean_distance.pyx"],
            include_dirs=include_dirs,
            **build_args,
        ),
        Extension(
            "wildboar.distance._dtw_distance",
            sources=["src/wildboar/distance/_dtw_distance.pyx"],
            include_dirs=include_dirs,
            **build_args,
        ),
        Extension(
            "wildboar.tree._tree_builder",
            sources=[
                "src/wildboar/tree/_tree_builder.pyx",
            ],
            include_dirs=include_dirs,
            **build_args,
        ),
        Extension(
            "wildboar.tree._impurity",
            sources=[
                "src/wildboar/tree/_impurity.pyx",
            ],
            include_dirs=include_dirs,
            **build_args,
        ),
    ]
    setup(ext_modules=cythonize(ext_modules))



# This file is part of wildboar
#
# wildboar is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# wildboar is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
# Authors: Isak Samsten

import math
import numbers
import warnings
from abc import ABCMeta, abstractmethod

import numpy as np
from sklearn import clone
from sklearn.cluster import KMeans, DBSCAN, OPTICS
from sklearn.ensemble import RandomForestClassifier
from sklearn.metrics import pairwise_distances, confusion_matrix
from sklearn.neighbors import NearestNeighbors
from sklearn.utils import check_random_state, check_array
from sklearn.utils.validation import check_is_fitted

from wildboar.linear_model import KernelLogisticRegression

__all__ = [
    "OutlierLabeler",
    "KMeansLabeler",
    "DensityLabeler",
    "MinorityLabeler",
    "MajorityLabeler",
    "EmmottLabeler",
]


class OutlierLabeler(metaclass=ABCMeta):
    """Base-class for outlier labelers"""

    @abstractmethod
    def fit(self, x, y=None):
        """Fit the outlier labeler to the give samples

        Parameters
        ----------
        x : array-like of shape (n_samples, n_timestep)
            The time series samples

        y : array-like of shape (n_samples, ) optional
            The optional original labels
        """
        pass

    @abstractmethod
    def transform(self, x, y=None):
        """Transform the labels of (a subset) of samples in x to inlier and outliers

        Parameters
        ----------
        x : array-like of shape (n_samples, n_timestep)
            The time series samples
        y : array-like of shape (n_samples, ), optional
            The optional original labels

        Returns
        -------
        x_new : array-like of shape (n_samples_new, n_timestep)
            The outlier and inlier samples

        y_new : array-like of shape (n_samples_new, )
            The labels
        """
        pass

    def fit_transform(self, x, y=None):
        return self.fit(x, y).transform(x, y)


class KMeansLabeler(OutlierLabeler):
    """KMeans labeler that assign an outlier label to the most deviating cluster

    Attributes
    ----------
    k_means_ : object
        The estimator for assigning points to the outlier class

    outlier_cluster_ : int
        The cluster index that is considered as outlier

    Warnings
    --------
    The implementation does not yet work as expected.
    """

    def __init__(self, *, n_clusters=None, n_outliers=None, random_state=None):
        """Construct a new labeler

        Parameters
        ----------
        n_clusters : int, optional
            Number of clusters to fit

        n_outliers : int or float, optional
            The number of outliers in the resulting dataset. This is not guaranteed.

        random_state : RandomState or int, optional
            The pseudo random state to ensure consistent results.
        """
        self.n_clusters = n_clusters
        self.n_outliers = n_outliers
        self.random_state = random_state

    def fit(self, x, y=None):
        n_clusters = self.n_clusters or max(5, x.shape[0] * 0.1)
        self.k_means_ = KMeans(n_clusters=n_clusters, random_state=self.random_state)

        self.k_means_.fit(x)
        _, count = np.unique(self.k_means_.labels_, return_counts=True)

        # skip clusters with too few samples
        invalid_clusters = np.where(count < 5)

        centroid_distance = pairwise_distances(
            self.k_means_.cluster_centers_, metric="euclidean"
        )

        # skip self matches and invalid clusters
        centroid_distance[centroid_distance == 0] = np.nan
        centroid_distance[invalid_clusters] = np.nan

        if np.all(np.isnan(x)):
            raise ValueError("no valid clusters")

        # hide the warning for ignored clusters
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", category=RuntimeWarning)
            self.outlier_cluster_ = np.nanargmax(np.nanmean(centroid_distance, axis=1))
        return self

    def fit_transform(self, x, y=None):
        return self.fit(x, y).transform(x, y)

    def transform(self, x, y=None):
        random_state = check_random_state(self.random_state)
        if self.n_outliers is None:
            n_outliers = x.shape[0]
        elif isinstance(self.n_outliers, float):
            if not 0.0 < self.n_outliers <= 1.0:
                raise ValueError(
                    "n_outliers must be in (0, 1], got %r" % self.n_outliers
                )
            n_outliers = math.ceil(x.shape[0] * self.n_outliers)
        elif isinstance(self.n_outliers, int):
            if not 0 < self.n_outliers < x.shape[0]:
                raise ValueError(
                    "n_outliers must be in (0, %d), got %r"
                    % (x.shape[0], self.n_outliers)
                )
            n_outliers = self.n_outliers
        else:
            raise ValueError("n_outliers (%s) not supported" % self.n_outliers)
        labels = self.k_means_.predict(x)
        outlier_indices = np.where(labels == self.outlier_cluster_)[0]
        inliers_indices = np.where(labels != self.outlier_cluster_)[0]

        idx = np.arange(outlier_indices.shape[0])
        if outlier_indices.shape[0] > n_outliers:
            random_state.shuffle(idx)
            outlier_indices = outlier_indices[idx[0:n_outliers]]
        y_new = np.ones(x.shape[0])
        y_new[outlier_indices] = -1
        return (
            np.concatenate([x[inliers_indices, :], x[outlier_indices, :]], axis=0),
            np.concatenate([y_new[inliers_indices], y_new[outlier_indices]], axis=0),
        )


DENSITY_ESTIMATORS = {"dbscan": DBSCAN(), "optics": OPTICS()}
DENSITY_ESTIMATORS_PARAMS = {}


class DensityLabeler(OutlierLabeler):
    """Density based clustering labeler

    Labels samples as outliers if a density cluster algorithm fail to assign them to a cluster
    """

    def __init__(self, *, estimator=None, estimator_params=None):
        self.estimator = estimator
        self.estimator_params = estimator_params

    def fit(self, x, y=None):
        if self.estimator is None:
            self.estimator_ = DENSITY_ESTIMATORS["dbscan"]
        elif self.estimator in DENSITY_ESTIMATORS:
            self.estimator_ = DENSITY_ESTIMATORS[self.estimator]
        else:
            self.estimator_ = self.estimator

        estimator_params = (
            self.estimator_params or DENSITY_ESTIMATORS_PARAMS[self.estimator]
            if self.estimator in DENSITY_ESTIMATORS_PARAMS
            else {}
        )
        self.estimator_.set_params(**estimator_params)
        self.estimator_.fit(x, y)
        label, count = np.unique(self.estimator_.labels_, return_counts=True)
        if len(count) == 1:
            raise ValueError("only a single cluster was formed")
        elif not np.any(label == -1):
            raise ValueError("no outlier points")

        return self

    def fit_transform(self, x, y=None):
        self.fit(x, y)
        y = np.ones(x.shape[0])
        y[self.estimator_.labels_ == -1] = -1
        return x, y

    def transform(self, x, y=None):
        y = np.ones(x.shape[0])
        if hasattr(self.estimator_, "predict"):
            y[self.estimator.predict(x) == -1] = -1
        else:
            raise ValueError("estimator does not support predict")


class MajorityLabeler(OutlierLabeler):
    """Labels the majority class as inliers

    Attributes
    ----------

    outlier_labels_ : ndarray
        The outlier labels
    """

    def __init__(self, n_outliers=None, random_state=None):
        """Create a new majority labeler

        Parameters
        ----------
        n_outliers : float, optional
            The fraction of outliers, by default None
        random_state : int or RandomState, optional
            The psudo-random number generator, by default None
        """
        self.n_outliers = n_outliers
        self.random_state = random_state

    def fit(self, x, y=None):
        x = check_array(x)
        y = check_array(y, ensure_2d=False)
        labels, counts = np.unique(y, return_counts=True)
        max_label = np.argmax(counts)
        self.outlier_label_ = labels[labels != labels[max_label]]
        return self

    def transform(self, x, y=None):
        check_is_fitted(self)
        random_state = check_random_state(self.random_state)
        outlier_indicator = np.isin(y, self.outlier_label_)
        outlier_indices = outlier_indicator.nonzero()[0]
        inlier_index = (~outlier_indicator).nonzero()[0]
        random_state.shuffle(outlier_indices)

        if self.n_outliers is None:
            n_outliers = outlier_indices.shape[0]
        elif isinstance(self.n_outliers, float):
            if not 0.0 < self.n_outliers <= 1.0:
                raise ValueError(
                    "n_outliers must be in (0, 1), got %r" % self.n_outliers
                )
            n_outliers = min(
                outlier_indices.shape[0],
                math.ceil(self.n_outliers * inlier_index.shape[0]),
            )
        else:
            raise ValueError("n_outliers (%r) is not supported" % self.n_outliers)

        x_outlier = x[outlier_indices[:n_outliers]]
        x_inlier = x[inlier_index]
        x = np.concatenate([x_outlier, x_inlier], axis=0)
        y = np.ones(x.shape[0])
        y[: x_outlier.shape[0]] = -1
        return x, y


class MinorityLabeler(OutlierLabeler):
    """Labels the minority class as the outlier

    Attributes
    ----------

    outlier_label_ : object
        The label of the outlier class
    """

    def __init__(self, n_outliers=None, random_state=None):
        self.n_outliers = n_outliers
        self.random_state = random_state

    def fit(self, x, y=None):
        labels, label_count = np.unique(y, return_counts=True)
        min_label = np.argmin(label_count)
        self.outlier_label_ = labels[min_label]
        return self

    def transform(self, x, y=None):
        random_state = check_random_state(self.random_state)
        outliers = np.where(y == self.outlier_label_)[0]
        random_state.shuffle(outliers)
        inliers = np.where(y != self.outlier_label_)[0]

        if self.n_outliers is None:
            n_outliers = outliers.shape[0]
        elif isinstance(self.n_outliers, numbers.Real):
            if not 0.0 < self.n_outliers <= 1.0:
                raise ValueError(
                    "n_outliers must be in (0, 1], got %r" % self.n_outliers
                )
            n_outliers = min(
                outliers.shape[0], math.ceil(self.n_outliers * inliers.shape[0])
            )
        else:
            raise ValueError("n_outliers (%r) is not supported" % self.n_outliers)

        n_outliers = min(n_outliers, x.shape[0])
        x = np.concatenate([x[outliers[0:n_outliers], :], x[inliers, :]], axis=0)
        y = np.concatenate([y[outliers[0:n_outliers]], y[inliers]])
        y[0:n_outliers] = -1
        y[n_outliers:] = 1
        return x, y


_DEFAULT_EMMOTT_SCALE = np.array([0, 0.16, 0.3, 0.5])


def _variation_dispersed(x, n_outliers, random_state):
    try:
        from sklearn_extra.cluster import KMedoids

        n_outliers = min(x.shape[0], n_outliers)
        f = KMedoids(n_clusters=n_outliers, random_state=random_state).fit(x)
        return f.medoid_indices_
    except ImportError:
        raise ValueError("variation (tight) require scikit-learn-extra.")


def _variation_tight(x, n_outliers, random_state):
    random_state = check_random_state(random_state)
    n_outliers = min(x.shape[0], n_outliers)
    if n_outliers == x.shape[0]:
        return np.arange(0, x.shape[0])
    sample = random_state.randint(0, x.shape[0])
    f = NearestNeighbors(n_neighbors=n_outliers).fit(x)
    return f.kneighbors(x[sample].reshape(1, -1), return_distance=False).reshape(-1)


def _set_random_states(estimator, random_state=None):
    random_state = check_random_state(random_state)
    to_set = {}
    for key in sorted(estimator.get_params(deep=True)):
        if key == "random_state" or key.endswith("__random_state"):
            to_set[key] = random_state.randint(np.iinfo(np.int32).max)

    if to_set:
        estimator.set_params(**to_set)


_EMMOTT_VARIATION = {"tight": _variation_tight, "dispersed": _variation_dispersed}


class EmmottLabeler(OutlierLabeler):
    """Create a synthetic outlier detection dataset from a labeled classification dataset
    using a method described by Emmott et.al. (2013).

    The Emmott labeler can reliably label both binary and multiclass datasets. For binary datasets
    a random label is selected as the outlier class. For multiclass datasets a set of classes with
    maximal confusion (as measured by ``confusion_estimator`` is selected as outlier label. For each
    outlier sample the ``difficulty_estimator`` assigns a difficulty score which is digitized into
    ranges and selected according to the ``difficulty`` parameters. Finally a sample of approximately
    ``n_outlier`` is selected either maximally dispersed or tight.

    Attributes
    ----------
    outlier_label_ : object
        The class or collection of classes used as outliers

    difficulty_estimator_ : object
        The estimator used to assess the difficulty of outlier samples

    confusion_estimator_ : object
        The estimator used to asses the class confusion (only if n_classes > 2)

    n_classes_ : int
        The number of classes

    Notes
    -----
    - For multiclass datasets the Emmott labeler require the package `networkx`
    - For dispersed outlier selection the Emmott labeler require the package `scikit-learn-extra`

    The difficulty parameters 'simplest' and 'hardest' are not described by Emmott et.al. (2013)

    Warnings
    --------
    n_outliers
        The number of outliers returned is dependent on the difficulty setting and the available
        number of samples of the minority class. If the minority class does not contain sufficient
        number of samples of the desired difficulty, fewer than n_outliers may be returned.


    References
    ----------
    Emmott, A. F., Das, S., Dietterich, T., Fern, A., & Wong, W. K. (2013).
        Systematic construction of anomaly detection benchmarks from real data.
        In Proceedings of the ACM SIGKDD workshop on outlier detection and description (pp. 16-21).

    """

    def __init__(
        self,
        n_outliers=None,
        *,
        confusion_estimator=None,
        difficulty_estimator=None,
        difficulty="simplest",
        scale=None,
        variation="tight",
        random_state=None,
    ):
        """Construct a new emmott labeler for synthetic outlier datasets

        Parameters
        ----------
        n_outliers : int, float, optional
            Number of desired (but not guaranteed) outliers in the resulting transformation.

        confusion_estimator : object, optional
            Estimator of class confusion for datasets where n_classes > 2. Default to a random forest classifier.

        difficulty_estimator : object, optional
            Estimator for sample difficulty. The difficulty estimator must support ``predict_proba``. Defaults
            to a kernel logistic regression model with a RBF-kernel.

        difficulty : {'any', 'simplest', 'hardest'}, int or array-like, optional
            The difficulty of the outlier points quantized according to scale. The value should be in the range
            `[1, len(scale)]` with lower difficulty denoting simpler outliers. If an array is given, multiple
            difficulties can be included, e.g., `[1, 4]` would mix easy and difficult outliers.

            - if 'any' outliers are sampled from all scores
            - if 'simplest' the simplest n_outliers are selected
            - if 'hardest' the hardest n_outliers are selected

        scale : array-like, optional
            The scale of quantized difficulty scores. Defaults to [0, 0.16, 0.3, 0.5]. Scores (which are probabilities
            in the range [0, 1]) are fit into the ranges using ``np.digitize(difficulty, scale)``.

        variation : {'tight', 'dispersed'}, optional
            Selection procedure for sampling outlier samples

            - if 'tight' a pivot point is selected and the ``n_outlier`` closest samples are selected according to
              their euclidean distance
            - if 'dispersed' ``n_outlier`` points are selected according to a facility location algorithm such that
              they are distributed among the outliers.

        random_state : RandomState or int, optinal
            A pseudo-random number generator to control the randomness of the algorithm.
        """
        self.n_outliers = n_outliers
        self.confusion_estimator = confusion_estimator
        self.difficulty_estimator = difficulty_estimator
        self.difficulty = difficulty
        self.scale = scale or _DEFAULT_EMMOTT_SCALE
        self.variation = variation
        self.random_state = random_state

    def fit(self, x, y=None):
        y = check_array(y, ensure_2d=False)
        x = check_array(x)
        random_state = check_random_state(self.random_state)
        self.n_classes_ = np.unique(y).shape[0]

        if self.n_classes_ > 2:
            if self.confusion_estimator is None:
                self.confusion_estimator_ = RandomForestClassifier(
                    n_jobs=-1, oob_score=True
                )
            else:
                self.confusion_estimator_ = clone(self.confusion_estimator)
            _set_random_states(self.confusion_estimator_, random_state)

            self.confusion_estimator_.fit(x, y)
            try:
                import networkx as nx

                if hasattr(self.confusion_estimator_, "oob_decision_function_"):
                    y_pred = self.confusion_estimator_.classes_[
                        np.argmax(
                            self.confusion_estimator_.oob_decision_function_, axis=1
                        )
                    ]
                else:
                    y_pred = self.confusion_estimator_.predict(x)
                cm = confusion_matrix(y, y_pred)  # TODO: use probabilities
                graph = nx.Graph()
                classes = self.confusion_estimator_.classes_
                graph.add_nodes_from(classes)
                for i in range(self.n_classes_):
                    for j in range(self.n_classes_):
                        if i != j:
                            graph.add_edge(
                                classes[i], classes[j], weight=cm[i][j] + cm[j][i]
                            )

                max_spanning_tree = nx.maximum_spanning_tree(graph, algorithm="kruskal")
                coloring = nx.algorithms.bipartite.color(max_spanning_tree)
                labeling = {1: [], 0: []}
                for cls in classes:
                    labeling[coloring[cls]].append(cls)

                zero = np.isin(y, labeling[0])
                one = np.isin(y, labeling[1])

                if np.sum(zero) <= np.sum(one):
                    self.outlier_label_ = np.array(labeling[0])
                else:
                    self.outlier_label_ = np.array(labeling[1])
            except ImportError:
                raise ValueError("for n_classes>2, `networkx` is required.")
        elif self.n_classes_ == 2:
            labels, counts = np.unique(y, return_counts=True)
            self.outlier_label_ = labels[np.argmin(counts)]
        else:
            raise ValueError("require more than 1 labels, got %r" % self.n_classes_)

        y_new = np.ones(x.shape[0], dtype=np.int)
        y_new[np.isin(y, self.outlier_label_)] = -1

        if self.difficulty_estimator is None:
            self.difficulty_estimator_ = KernelLogisticRegression(
                kernel="poly", max_iter=1000
            )
        else:
            self.difficulty_estimator_ = clone(self.difficulty_estimator)
        _set_random_states(self.difficulty_estimator_, random_state)
        self.difficulty_estimator_.fit(x, y_new)
        return self

    def fit_transform(self, x, y=None):
        self.fit(x, y)
        if hasattr(self.difficulty_estimator_, "oob_decision_function_"):
            difficulty_estimate = self.difficulty_estimator_.oob_decision_function_
        else:
            difficulty_estimate = self.difficulty_estimator_.predict_proba(x)
        difficulty_estimate = difficulty_estimate[
            :, np.where(self.difficulty_estimator_.classes_ == 1)[0]
        ]
        return self._transform(x, y, difficulty_estimate)

    def transform(self, x, y=None):
        difficulty_estimate = self.difficulty_estimator_.predict_proba(x)
        difficulty_estimate = difficulty_estimate[
            :, np.where(self.difficulty_estimator_.classes_ == 1)[0]
        ]
        return self._transform(x, y, difficulty_estimate)

    def _transform(self, x, y, difficulty_estimate):
        y_new = np.ones(x.shape[0], dtype=np.int)
        y_new[np.isin(y, self.outlier_label_)] = -1
        y = y_new
        random_state = check_random_state(self.random_state)
        outliers_indices = np.where(y == -1)
        inlier_indices = np.where(y == 1)

        if self.variation in _EMMOTT_VARIATION:
            variation = _EMMOTT_VARIATION[self.variation]
        else:
            raise ValueError("variation (%s) is not supported" % self.variation)

        x_outliers = x[outliers_indices]
        y_outliers = y[outliers_indices]
        difficulty_estimate = difficulty_estimate[outliers_indices].reshape(-1)

        if self.n_outliers is None:
            n_outliers = y_outliers.shape[0]
        elif isinstance(self.n_outliers, float):
            if not 0.0 < self.n_outliers <= 1.0:
                raise ValueError(
                    "n_outliers must be in (0, 1], got %r" % self.n_outliers
                )
            n_outliers = min(
                y_outliers.shape[0],
                math.ceil(self.n_outliers * inlier_indices[0].shape[0]),
            )
        elif isinstance(self.n_outliers, int):
            if 0 < self.n_outliers <= y_outliers.shape[0]:
                raise ValueError(
                    "n_outliers must be in (0, %d], got %r"
                    % (y_outliers.shape[0], self.n_outliers)
                )
            n_outliers = self.n_outliers
        else:
            raise ValueError("n_outlier (%r) is not supported" % self.n_outliers)

        if n_outliers < y_outliers.shape[0]:
            difficulty_scores = np.digitize(difficulty_estimate, self.scale)
            if isinstance(self.difficulty, str):
                if self.difficulty == "any":
                    outlier_selector = np.arange(0, difficulty_estimate.shape[0])
                elif self.difficulty == "simplest":
                    outlier_selector = np.argpartition(difficulty_estimate, n_outliers)[
                        :n_outliers
                    ]
                elif self.difficulty == "hardest":
                    outlier_selector = np.argpartition(
                        difficulty_estimate, -n_outliers
                    )[-n_outliers:]
                else:
                    raise ValueError(
                        "difficulty (%s) is not supported" % self.difficulty
                    )
            else:
                outlier_selector = np.isin(difficulty_scores, self.difficulty)
            x_outliers = x_outliers[outlier_selector]
            if x_outliers.shape[0] == 0:
                scores, counts = np.unique(difficulty_scores, return_counts=True)
                raise ValueError(
                    "no samples with the requested difficulty %s, available %s"
                    % (
                        self.difficulty,
                        ", ".join(["%d: %d" % (s, c) for s, c in zip(scores, counts)]),
                    )
                )

        outlier_sampled = variation(
            x_outliers, n_outliers, random_state.randint(np.iinfo(np.int32).max)
        )

        return (
            np.concatenate([x[inlier_indices], x_outliers[outlier_sampled]], axis=0),
            np.concatenate([y[inlier_indices], y_outliers[outlier_sampled]], axis=0),
        )



# cython: language_level=3

# This file is part of wildboar
#
# wildboar is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# wildboar is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

# Authors: Isak Samsten

cdef class RollingVariance:
    cdef double _n_samples
    cdef double _m
    cdef double _s
    cdef double _sum

    cdef void _reset(self) nogil

    cdef void _add(self, double weight, double value) nogil

    cdef void _remove(self, double weight, double value) nogil

    cdef double _mean(self) nogil

    cdef double _variance(self) nogil

cdef enum:
    RAND_R_MAX = 2147483647

cdef void print_c_array_d(object name, double *arr, Py_ssize_t length)

cdef void print_c_array_i(object name, Py_ssize_t *arr, Py_ssize_t length)

cdef Py_ssize_t label_distribution(const Py_ssize_t *samples, double *sample_weights, Py_ssize_t start, Py_ssize_t end,
                               const Py_ssize_t *labels, Py_ssize_t labels_stride, Py_ssize_t n_classes,
                               double*n_weighted_samples, double*dist) nogil

cdef void argsort(double *values, Py_ssize_t *order, Py_ssize_t length) nogil

cdef size_t rand_r(size_t *seed) nogil

cdef size_t rand_int(size_t min_val, size_t max_val, size_t *seed) nogil

cdef double rand_uniform(double low, double high, size_t *random_state) nogil

cdef int realloc_array(void** a, Py_ssize_t p, Py_ssize_t size, Py_ssize_t *cap)  nogil except -1

cdef int safe_realloc(void** ptr, Py_ssize_t new_size) nogil except -1

cdef void fast_mean_std(Py_ssize_t offset, Py_ssize_t stride, Py_ssize_t length, double* data, double *mean, double* std) nogil